smin <- c(## floor = 0.15,
          ## ceiling = 0.15,
          sqrt = 0.2,
          ## sign = 0.15,
          exp = 0.44,
          expm1 = 0.41,
          log1p = 0.45,
          sin = 0.42,
          cos = 0.46,
          tan = 0.68,
          acos = 0.44,
          asin = 0.47,
          lgamma = 1.98,
          gamma = 1.4,
          digamma = 4.61,
          trigamma = 13.5,
          gammaCody = 0.48,
          lbeta = 4.67,
          beta = 4.18,
          lchoose = 6.51,
          choose = 0.96,
          dchisq = 0.3,
          pchisq = 0.35,
          qchisq = 15.81,
          dexp = 0.23,
          pexp = 0.24,
          qexp = 0.7,
          dgeom = 1.48,
          pgeom = 1.34,
          qgeom = 1.15,
          dpois = 1.77,
          ppois = 3.42,
          qpois = 7.54,
          dt = 2.32,
          pt = 10.42,
          qt = 27.14,
          psigamma = 25.21,
          dbeta = 3.95,
          pbeta = 8.81,
          qbeta = 52.06,
          dbinom = 4.21,
          pbinom = 7.03,
          qbinom = 15.41,
          dcauchy = 0.47,
          pcauchy = 1.03,
          qcauchy = 1.05,
          df = 1.71,
          pf = 7.48,
          qf = 38.59,
          dgamma = 2.57,
          pgamma = 4.65,
          qgamma = 15.9,
          dlnorm = 1.36,
          plnorm = 1.64,
          qlnorm = 1.33,
          dlogis = 0.98,
          plogis = 0.9,
          qlogis = 0.9,
          dnbinom = 4.4,
          pnbinom = 7.07,
          qnbinom = 16.35,
          dnorm = 1,
          pnorm = 0.8,
          qnorm = 0.8,
          dunif = 0.4,
          punif = 0.54,
          qunif = 0.49,
          dweibull = 1.13,
          pweibull = 1.03,
          qweibull = 1.48,
          dnchisq = 3.51,
          pnchisq = 456.42,
          qnchisq = 16014.39,
          dnt = 32.91,
          pnt = 11.86,
          qnt = 850.57,
          dhyper = 8.63,
          phyper = 8.15,
          qhyper = 15.72,
          dnbeta = 9.82,
          pnbeta = 19.48,
          qnbeta = 1065.76,
          dnf = 10.37,
          pnf = 15.7,
          qnf = 941.92,
          ptukey = 27144.18,
          qtukey = 141724.2)

parallelThresholds <- NULL

getNumProcs <- function() {
    n <- .Call("do_NumProcs")
    if (n >= 0) n else NA
}

getNumPnmathThreads <- function() .Call("do_GetPMAX")

setNumPnmathThreads <- function(n) {
    NCPU <- getNumProcs()
    if (is.finite(n) && (NCPU < 1 || n <= NCPU))
        .Call("do_SetPMAX", n)
    .Call("do_GetPMAX")
}

cutoffData <- new.env(parent = emptyenv())

cal <- function(fun, n, mkx, ..., N = 1000000) {
    sim <- function(n) {
        R <- ceiling(N / n)
        x <- mkx(n)
        1000 * system.time(for (i in 1:R) fun(x, ...))[3] / R
    }
    sapply(n, sim)
}

cal2 <- function(fun, n, mkx, ..., N = 5000, R = 10) {
    t1 <- matrix(0, nrow = length(n), ncol = R)
    t2 <- matrix(0, nrow = length(n), ncol = R)

    for (i in 1:R) {
        t1[,i] <- cal(fun, n, mkx, ..., N = N)
        t2[,i] <- cal(fun, n, mkx, ..., P = 2, N = N)
    }
    list(t1 = t1, t2 = t2, n = n)
}

cal.dnorm <- function(R = 10, N = 5000) {
    n <- c(8, 16, 64, 128, 256, 512, 1024)
    mkx <- function(n) (1 : n) / (n + 1)
    cal2(p.dnorm, n, mkx, 2, 4, R = R, N = N)
}

calibratePnmath <- function(m, N = 100000) {
    if (missing(m)) {
        dn <- cal.dnorm(N = N)
        dn2coef <- lm(apply(dn$t2,1,median)~dn$n)$coef
        dn1coef <- lm(apply(dn$t1,1,median)~dn$n)$coef
        m <- dn2coef[1] / dn1coef[2]
    }
    
    c2 <- ceiling(2 * m / smin)
    c4 <- ceiling(4 * m / smin)
    c8 <- ceiling(8 * m / smin)
    nm <- names(smin)
    for (i in seq_along(smin))
        .Call("do_vnmathEnter", nm[i], c2[i], c4[i], c8[i])

    assign("c2", c2, cutoffData)
    assign("m", m, cutoffData)
    m
}    

getParallelThresholds <- function()
    get("c2", cutoffData)

.onLoad <- function(libname, pkgname) {
    .Call("do_initRpartest")

    NCPU <- .Call("do_NumProcs")
    ev <- Sys.getenv("OMP_NUM_THREADS")
    if (ev != "") {
        n <- as.integer(ev)
        if (is.finite(n) && (NCPU < 1 || n < NCPU))
            .Call("do_SetPMAX", n)
    }

    .Call("do_replaceBuiltins")

    sn <- Sys.info()['sysname']
    if (sn == "Linux") calibratePnmath(200)
    else if (sn == "Darwin") calibratePnmath(500)
    else {
        require(stats)
        message("Calibrating ... ", appendLF = FALSE)
        calibratePnmath()
        message("done\n")
    }
}
